TASK_INSTRUCTIONS: dict[str, str] = {
    "etpc": "Paraphrase the provided text into a single paraphrase by using all paraphrase types.",
    "wmt19_de_en": "Translate the provided text from German to English.",
    "simple_ethical_questions": "Answer the provided question by choosing option A), B), C), or D). Include the letter corresponding to your answer in the solution.",
    "squad_v2": "Answer the following question. If the question is not answerable with the provided information, write '[UNKNOWN]'.",
    "strategyqa": "Answer the following question with A) Yes or B) No. Include the letter corresponding to your answer in the solution.",
    "xsum": "Summarize the provided text into one sentence.",
    "mmlu_pro": "Answer the provided question by choosing one option. Include the letter corresponding to your answer in the solution.",
    "musr": "Solve the mystery. Make sure to always include the full name of the suspect in your solution.",
    "gpqa": "Answer the provided question by choosing option A), B), C), or D). Include the letter corresponding to your answer in the solution.",
    "math_lvl5": "Answer the following math question. Output the number as the final solution.",
    "ifeval": "Follow the instruction.",
    "winogrande": "Correctly fill the blank space _ by choosing option A) or B). Include the letter corresponding to your answer in the solution.",
    "ethics": "Would most people believe this reasonable or unreasonable to say? Answer with A) Reasonable or B) Unreasonable. Include the letter corresponding to your answer in the solution.",
    "aqua_rat": "Answer the following math question. Include the letter corresponding to your answer in the solution."
}
